

Millloading = {};

function Millloading.prerequisitesPresent(specializations)
	    return SpecializationUtil.hasSpecialization(Fillable, specializations) and SpecializationUtil.hasSpecialization(FillVolume, specializations);
end;

function Millloading:load(savegame)

    self.setOverloadingActive2           = Millloading.setOverloadingActive2;
    self.doOverload2                    = Millloading.doOverload2;
    self.getIsFoldAllowed2               = Utils.overwrittenFunction(self.getIsFoldAllowed, Millloading.getIsFoldAllowed2);
    self.getPtoRpm2                      = Utils.overwrittenFunction(self.getPtoRpm, Millloading.getPtoRpm2);
    self.getDoConsumePtoPower2           = Utils.overwrittenFunction(self.getDoConsumePtoPower, Millloading.getDoConsumePtoPower2);
    self.getIsOverloadingAllowed2        = Millloading.getIsOverloadingAllowed2;
    self.findValidTrailer2               = Millloading.findValidTrailer2;
    self.loadTrailerTriggers2            = Millloading.loadTrailerTriggers2;
    self.onOverloadingTrailerTrigger2    = Millloading.onOverloadingTrailerTrigger2;
    self.getAutomaticOverloadingState2   = Millloading.getAutomaticOverloadingState2;
    self.getOverloadingActive2 = Utils.overwrittenFunction(self.getOverloadingActive, Millloading.getOverloadingActive2);

    self.startOverloadingDelay2          = Millloading.startOverloadingDelay2;
    self.stopOverloadingDelay2           = Millloading.stopOverloadingDelay2;
    self.updateOverloadingDelay2         = Millloading.updateOverloadingDelay2;
    self.setOverloadingDelay2            = Millloading.setOverloadingDelay2;
    self.getOverloadingDelayIsActive2    = Millloading.getOverloadingDelayIsActive2;
    self.getAllowDelayedUnloading2       = Millloading.getAllowDelayedUnloading2;
    self.getAllowDelayEffects2           = Millloading.getAllowDelayEffects2;
    self.getAllowDelayedEffects2         = Millloading.getAllowDelayedEffects2;

    Utils.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.combineTrailerTriggers.combineTrailerTrigger", "vehicle.overloading.trailerTriggers.trailerTrigger")
    Utils.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.overloadingCapacity", "vehicle.overloading#capacity")
    Utils.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.trailerTriggers#foldMinTime", "vehicle.overloading#foldMinTime")
    Utils.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.trailerTriggers#foldMaxTime", "vehicle.overloading#foldMaxTime")
    Utils.checkDeprecatedXMLElements(self.xmlFile, self.configFileName, "vehicle.trailerTriggers#allowFoldWhileOverloading", "vehicle.overloading#allowFoldWhileOverloading")

    self.overloading = {};
    self.overloading.trailerTriggers = {};
    self:loadTrailerTriggers2(self.overloading.trailerTriggers, self.xmlFile);
    for _, trigger in pairs(self.overloading.trailerTriggers) do
        addTrigger(trigger.node, "onOverloadingTrailerTrigger2", self);
    end
    self.overloading.trailersInRange = {};
    self.overloading.capacity = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.overloading#capacity"), 10);
    self.overloading.fillUnitIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.overloading#fillUnitIndex"), 1);
    self.overloading.unloadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.overloading#unloadInfoIndex"), 1);
    self.overloading.loadInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.overloading#loadInfoIndex"), 1);
    self.overloading.dischargeInfoIndex = Utils.getNoNil(getXMLInt(self.xmlFile, "vehicle.overloading#dischargeInfoIndex"), 1);
    self.overloading.stopOverloadIfEmpty = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.overloading#stopOverloadIfEmpty"), true);
    self.overloading.forceActive = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.overloading#forceActive"), false);
    self.overloading.canToggleOverloading = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.overloading#canToggleOverloading"), true);
    self.overloading.foldMinTime = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.overloading#foldMinTime"), 0);
    self.overloading.foldMaxTime = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.overloading#foldMaxTime"), 1);
    self.overloading.allowFoldWhileOverloading = Utils.getNoNil(getXMLBool(self.xmlFile, "vehicle.overloading#allowFoldWhileOverloading"), false);
    self.overloading.isActive = false;
    self.overloading.didOverload = false;

    self.overloading.delay = {};
    self.overloading.delay.current = 0;
    self.overloading.delay.unloadStartTime = 0;
    self.overloading.delay.sentCurrent = 0;
    self.overloading.delay.sentUnloadStartTime = 0;
    self.overloading.delay.dirtyFlag = self:getNextDirtyFlag();
    self.overloading.delay.time = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.overloading#delay"), Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.pipe#delay"), 0)) * 1000;

    self.isSelectable = true;

    local linkNode = Utils.indexToObject(self.components, Utils.getNoNil(getXMLString(self.xmlFile, "vehicle.fillSound#linkNode"), "0>"));
    self.sampleFill = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.fillSound", nil, self.baseDirectory, linkNode);
    self.sampleFillStopTime = 0;


    self.overloadingDirtyFlag = self:getNextDirtyFlag();

    self:addConflictCheckedInput(InputBinding.TOGGLE_PIPE);
end;



function Millloading:delete()
    for _, trailerTrigger in pairs(self.overloading.trailerTriggers) do
        removeTrigger(trailerTrigger.node);
    end

    if self.isClient then
        SoundUtil.deleteSample(self.sampleFill);
    end
end;

function Millloading:mouseEvent(posX, posY, isDown, isUp, button)
end;
function Millloading:keyEvent(unicode, sym, modifier, isDown)
end;

function Millloading:readStream(streamId, connection)
    if connection:getIsServer() then
        self.overloading.didOverload = streamReadBool(streamId);
        self:setOverloadingActive2(streamReadBool(streamId), true);
    end
end;


function Millloading:writeStream(streamId, connection)
    if not connection:getIsServer() then
        streamWriteBool(streamId, self.overloading.didOverload);
        streamWriteBool(streamId, self.overloading.isActive);
    end
end;

function Millloading:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        if streamReadBool(streamId) then
            self.overloading.didOverload = streamReadBool(streamId);
        end;

        if streamReadBool(streamId) then
            self.overloading.delay.current = streamReadUIntN(streamId, 6)/63*self.overloading.delay.time;

            if streamReadBool(streamId) then
                self.overloading.delay.unloadStartTime = 1;
            else
                self.overloading.delay.unloadStartTime = -1;
            end;
        end;
    end
end;

function Millloading:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        if streamWriteBool(streamId, bitAND(dirtyMask, self.overloadingDirtyFlag) ~= 0) then
            streamWriteBool(streamId, self.overloading.didOverload);
        end;

        if streamWriteBool(streamId, bitAND(dirtyMask, self.overloading.delay.dirtyFlag) ~= 0) then
            streamWriteUIntN(streamId, Utils.clamp(math.floor(self.overloading.delay.current/self.overloading.delay.time*63), 0, 63), 6);
            streamWriteBool(streamId, self.overloading.delay.unloadStartTime > 0);
        end;
    end
end;

function Millloading:update(dt)
    if self.isClient and self:getIsActive() then
        if InputBinding.hasEvent(InputBinding.TOGGLE_TIPSTATE) and self.overloading.canToggleOverloading then
            if self:getIsActiveForInput(true) and not self:hasInputConflictWithSelection() then
                if self:getIsOverloadingAllowed2() then
                    self:setOverloadingActive2(not self.overloading.isActive);
                end
            end
        end
    end
end;

function Millloading:updateTick(dt)
    -- sound for a more holistic refactoring
    if self.isClient then
        if self.overloading.didOverload or self.isFilling then
            self.sampleFillStopTime = g_currentMission.time + 150;
        end

        if self.sampleFillStopTime > g_currentMission.time then
            if self:getIsActiveForSound(true) then
                SoundUtil.playSample(self.sampleFill, 0, 0, nil);
                SoundUtil.stop3DSample(self.sampleFill);
            else
                SoundUtil.stopSample(self.sampleFill);
                SoundUtil.play3DSample(self.sampleFill);
            end
        else
            SoundUtil.stopSample(self.sampleFill);
            SoundUtil.stop3DSample(self.sampleFill);
        end
    end

    if self.isServer then
        self:updateOverloadingDelay2(dt);

        if self.overloading.delay.current ~= self.overloading.delay.sentCurrent or self.overloading.delay.unloadStartTime ~= self.overloading.delay.sentUnloadStartTime then
            self:raiseDirtyFlags(self.overloading.delay.dirtyFlag);

            self.overloading.delay.sentCurrent = self.overloading.delay.current;
            self.overloading.delay.sentUnloadStartTime = self.overloading.delay.unloadStartTime;
        end;

        if self:getUnitFillLevel(self.overloading.fillUnitIndex) <= 0 and self:getOverloadingDelayIsActive2()  then
            self:stopOverloadingDelay2();
        end;

        self.overloading.didOverload = false;
        if self:getIsActive() or self.overloading.forceActive or next(self.overloading.trailersInRange) ~= nil or self:getOverloadingActive2() then -- and self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0) then

            -- first update the target trailer
            self.targetTrailer = self:findValidTrailer2(self.overloading.trailersInRange);

            -- second, update status
            if self.overloading.isActive then
                if self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 and not self:getOverloadingDelayIsActive2() then
                    self:startOverloadingDelay2();
                end;

                if not self:getIsOverloadingAllowed2() then
                    self:setOverloadingActive2(false);
                else
                    if self.overloading.stopOverloadIfEmpty then
                        if self:getUnitFillLevel(self.overloading.fillUnitIndex) == 0 then
                            self:setOverloadingActive2(false);
                        end
                    end
                end

                -- third, overload if required
                if (self.targetTrailer ~= nil or self.shovelTargetFound ~= nil) and self:getAllowDelayedUnloading2() then

                    local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
                    local fillLevel = self:getUnitFillLevel(self.overloading.fillUnitIndex);
                    if self.lastLostFillLevel ~= nil then
                        fillLevel = math.max(fillLevel, self.lastLostFillLevel);
                    end

                    local deltaFill = fillLevel;
                    if self:getUnitCapacity(self.overloading.fillUnitIndex) > 0 then
                        deltaFill = math.min(fillLevel, self.overloading.capacity*dt/1000.0);
                    end

                    if self.targetTrailer ~= nil then
                        self.overloading.didOverload = self:doOverload2(self.targetTrailer, fillType, deltaFill);
                    elseif self.shovelTargetFound ~= nil then
                        local realDelta = self.shovelTargetFound:addPipeFillLevel(self, deltaFill, fillType);
                        if realDelta > 0 then
                            self:setUnitFillLevel(self.overloading.fillUnitIndex, self:getUnitFillLevel(self.overloading.fillUnitIndex) - realDelta, fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
                            self.overloading.didOverload = true;
                        else
                            self:setOverloadingActive2(false);
                        end;
                    end;
                end
            else
                if self:getAutomaticOverloadingState2() then
                    self:setOverloadingActive2(true);
                end
            end
        end

        if self.overloading.didOverload ~= self.sendDidOverload then
            self:raiseDirtyFlags(self.overloadingDirtyFlag);
            self.sendDidOverload = self.overloading.didOverload;
        end
    end
end;


function Millloading:draw()
    if self.isClient and self:getIsActiveForInput(true) and not self:hasInputConflictWithSelection() then
        if self.overloading.canToggleOverloading and self:getIsOverloadingAllowed2() then
            if self.overloading.isActive then
                g_currentMission:addHelpButtonText(g_i18n:getText("action_stopOverloading"), InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_HIGH);
            else
                if self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 and self.targetTrailer ~= nil then
                    g_currentMission:addHelpButtonText(g_i18n:getText("action_startOverloading"), InputBinding.TOGGLE_TIPSTATE, nil, GS_PRIO_HIGH);
                end
            end
        end
    end
end;


function Millloading:setOverloadingActive2(isActive, noEventSend)
    if isActive ~= self.overloading.isActive then
        OverloadingToggleUnloadEvent2.sendEvent(self, isActive, noEventSend);
        self.overloading.isActive = isActive;

        if not isActive then
            self.pipeUnloadingStartTime = -1;
        end;
    end
end;


function Millloading:getOverloadingActive2(superFunc)
    if self.overloading.isActive then
        return true;
    end;

    if superFunc ~= nil then
        return superFunc(self);
    end
    return false;
end;


function Millloading:doOverload2(targetTrailer, fillType, deltaFill)
	millfillType=FillUtil.FILLTYPE_CCMRAW;
    targetTrailer:resetFillLevelIfNeeded(millfillType);
print("Do Overloading")
    local trailerFillLevel = targetTrailer:getFillLevel(millfillType);
    local trailerCapacity = targetTrailer:getCapacity(millfillType);

    deltaFill = math.min(deltaFill, trailerCapacity - trailerFillLevel);

    if deltaFill > 0 then
        targetTrailer:setFillLevel(trailerFillLevel+(deltaFill*1.2), millfillType, false, self.fillVolumeDischargeInfos[self.overloading.dischargeInfoIndex]);
        self:setFillLevel(self:getUnitFillLevel(self.overloading.fillUnitIndex)-deltaFill, fillType, false, self.fillVolumeUnloadInfos[self.overloading.unloadInfoIndex]);
        return true;
    else
        return false;
    end
end;


function Millloading:getIsFoldAllowed2(superFunc, onAiTurnOn)
    if not self.overloading.allowFoldWhileOverloading and self.overloading.isActive then
        if self:getUnitFillLevel(self.overloading.fillUnitIndex) > 0 then
            return false;
        end
    end
    if self.foldAnimTime ~= nil and (self.foldAnimTime > self.overloading.foldMaxTime or self.foldAnimTime < self.overloading.foldMinTime) then
        return false;
    end

    if superFunc ~= nil then
        return superFunc(self, onAiTurnOn);
    end
    return true;
end;


function Millloading:getPtoRpm2(superFunc)
    local ptoRpm = 0;
    if superFunc ~= nil then
        ptoRpm = superFunc(self);
    end
    if self.overloading.didOverload and self.powerConsumer ~= nil then
        ptoRpm = math.max(ptoRpm, self.powerConsumer.ptoRpm);
    end
    return ptoRpm;
end;


function Millloading:getDoConsumePtoPower2(superFunc)
    local doConsumePower = superFunc(self);
    return self.overloading.didOverload or doConsumePower;
end;

function Millloading:getIsOverloadingAllowed2()
    if self.foldAnimTime ~= nil then
        if self.foldAnimTime > self.overloading.foldMaxTime or self.foldAnimTime < self.overloading.foldMinTime then
            return false;
        end
    end
    return true;
end;

function Millloading:findValidTrailer2(trailers)
millfillType=FillUtil.FILLTYPE_CCMRAW;
    local fillType = self:getUnitLastValidFillType(self.overloading.fillUnitIndex);
    for trailer,value in pairs(trailers) do
        if not entityExists(trailer.components[1].node) then
            value = 0;
        end
        if value > 0 then
            if trailer:allowFillType(millfillType) then
                if trailer:getFillLevel(millfillType) < trailer:getCapacity(millfillType) then
                    return trailer;
                end
            end
        end
    end
end;


function Millloading:getAutomaticOverloadingState2()
    return false
end;

function Millloading:loadTrailerTriggers2(triggers, xmlFile)
    local i = 0;
    while true do
        local key = string.format("vehicle.overloading.trailerTriggers.trailerTrigger(%d)", i);
        if not hasXMLProperty(self.xmlFile, key) then
            break;
        end
        local node = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#index"));
        if node ~= nil then
            triggers[node] = {node=node};
        end
        i = i + 1;
    end
end;


function Millloading:onOverloadingTrailerTrigger2(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        local trailer = g_currentMission.nodeToVehicle[otherId];
        if trailer ~= nil and trailer.fillRootNode ~= nil and trailer ~= self then
            if onEnter then
                self.overloading.trailersInRange[trailer] = Utils.getNoNil(self.overloading.trailersInRange[trailer], 0) + 1;
            else
                self.overloading.trailersInRange[trailer] = math.max(Utils.getNoNil(self.overloading.trailersInRange[trailer], 0) - 1, 0);
                if self.overloading.trailersInRange[trailer] == 0 then
                    self.overloading.trailersInRange[trailer] = nil
                end
            end
        end
    end
end;


function Millloading:startOverloadingDelay2()
    self.overloading.delay.unloadStartTime = g_currentMission.time;
    self.overloading.delay.current = 0;
end;


function Millloading:stopOverloadingDelay2()
    self.overloading.delay.unloadStartTime = -1;
end;


function Millloading:updateOverloadingDelay2(dt)
    if self.overloading.delay.unloadStartTime > 0 then
        if self.overloading.delay.current < self.overloading.delay.time then
            self:setOverloadingDelay2(math.min(self.overloading.delay.current + dt, self.overloading.delay.time));
        end;
    else
        if self.overloading.delay.current > 0 then
            self:setOverloadingDelay2(math.max(self.overloading.delay.current - dt, 0));
        end;
    end;
end;


function Millloading:setOverloadingDelay2(delay)
    self.overloading.delay.current = delay;
end;


function Millloading:getOverloadingDelayIsActive2()
    return self.overloading.delay.unloadStartTime > 0;
end;


function Millloading:getAllowDelayEffects2()
    if self:getUnitCapacity(self.overloading.fillUnitIndex) == 0 then
        return self.overloading.didOverload;
    end;

    return self.overloading.delay.current > 0 and self:getOverloadingDelayIsActive2();
end;


function Millloading:getAllowDelayedEffects2()
    if self:getUnitCapacity(self.overloading.fillUnitIndex) == 0 then
        return self.overloading.didOverload;
    end;

    return (self.overloading.delay.current > 0 and not self:getOverloadingDelayIsActive2()) or self:getAllowDelayedUnloading2();
end;


function Millloading:getAllowDelayedUnloading2()
    return self.overloading.delay.current == self.overloading.delay.time;
end;



OverloadingToggleUnloadEvent2 = {};
OverloadingToggleUnloadEvent2_mt = Class(OverloadingToggleUnloadEvent2, Event);

InitEventClass(OverloadingToggleUnloadEvent2, "OverloadingToggleUnloadEvent2");

function OverloadingToggleUnloadEvent2:emptyNew()
    local self = Event:new(OverloadingToggleUnloadEvent2_mt);
    return self;
end;


function OverloadingToggleUnloadEvent2:new(vehicle, isActive)
    local self = OverloadingToggleUnloadEvent2:emptyNew()
    self.vehicle = vehicle;
    self.isActive = isActive;
    return self;
end;

function OverloadingToggleUnloadEvent2:readStream(streamId, connection)
    self.vehicle = readNetworkNodeObject(streamId);
    self.isActive = streamReadBool(streamId);
    self:run(connection);
end;

function OverloadingToggleUnloadEvent2:writeStream(streamId, connection)
    writeNetworkNodeObject(streamId, self.vehicle);
    streamWriteBool(streamId, self.isActive);
end;

function OverloadingToggleUnloadEvent2:run(connection)
    self.vehicle:setOverloadingActive2(self.isActive, true);
    if not connection:getIsServer() then
        g_server:broadcastEvent(OverloadingToggleUnloadEvent2:new(self.vehicle, self.isActive), nil, connection, self.vehicle);
    end;
end;


function OverloadingToggleUnloadEvent2.sendEvent(vehicle, isActive, noEventSend)
    if noEventSend == nil or noEventSend == false then
        if g_server ~= nil then
            g_server:broadcastEvent(OverloadingToggleUnloadEvent2:new(vehicle, isActive), nil, nil, vehicle);
        else
            g_client:getServerConnection():sendEvent(OverloadingToggleUnloadEvent2:new(vehicle, isActive));
        end;
    end;
end;